import { __decorate } from "tslib";
import { cssDirective, FASTElement, Observable, SubscriberSet, } from "@microsoft/fast-element";
import { composedContains, composedParent } from "@microsoft/fast-element/utilities";
import { PropertyTargetManager, RootStyleSheetTarget, } from "./custom-property-manager.js";
import { DesignTokenNode, } from "./core/design-token-node.js";
/**
 * @public
 */
export class DesignToken {
    constructor(configuration) {
        this.subscriberNotifier = {
            handleChange: (source, change) => {
                const record = {
                    target: change.target === FASTDesignTokenNode.defaultNode
                        ? "default"
                        : change.target.target,
                    token: this,
                };
                this.subscribers.notify(record);
            },
        };
        this.name = configuration.name;
        Observable.getNotifier(this).subscribe(this.subscriberNotifier);
    }
    /**
     * The default value of the token (alias of {@link DesignToken.default})
     */
    get $value() {
        return this.default;
    }
    /**
     * The default value of the token, or undefined if it has not been set.
     */
    get default() {
        return FASTDesignTokenNode.defaultNode.getTokenValue(this);
    }
    get subscribers() {
        if (this._subscribers) {
            return this._subscribers;
        }
        this._subscribers = new SubscriberSet(this);
        return this._subscribers;
    }
    static isCSSDesignTokenConfiguration(config) {
        return (typeof config.cssCustomPropertyName ===
            "string");
    }
    static create(config) {
        if (typeof config === "string") {
            return new CSSDesignToken({ name: config, cssCustomPropertyName: config });
        }
        else {
            return DesignToken.isCSSDesignTokenConfiguration(config)
                ? new CSSDesignToken(config)
                : new DesignToken(config);
        }
    }
    /**
     * Configures the strategy for resolving hierarchical relationships between FASTElement targets.
     */
    static withStrategy(strategy) {
        FASTDesignTokenNode.withStrategy(strategy);
    }
    /**
     * Registers and element or document as a DesignToken root.
     * {@link CSSDesignToken | CSSDesignTokens} with default values assigned via
     * {@link DesignToken.withDefault} will emit CSS custom properties to all
     * registered roots.
     * @param target - The root to register
     */
    static registerRoot(target = document) {
        if (target instanceof FASTElement || target instanceof Document) {
            target = PropertyTargetManager.getOrCreate(target);
        }
        RootStyleSheetTarget.registerRoot(target);
    }
    /**
     * Unregister an element or document as a DesignToken root.
     * @param target - The root to deregister
     */
    static unregisterRoot(target = document) {
        if (target instanceof FASTElement || target instanceof Document) {
            target = PropertyTargetManager.getOrCreate(target);
        }
        RootStyleSheetTarget.unregisterRoot(target);
    }
    /**
     * Retrieves the value of the token for a target element.
     */
    getValueFor(target) {
        return FASTDesignTokenNode.getOrCreate(target).getTokenValue(this);
    }
    /**
     * Sets the value of the token for a target element.
     */
    setValueFor(target, value) {
        FASTDesignTokenNode.getOrCreate(target).setTokenValue(this, this.normalizeValue(value));
    }
    /**
     * Deletes the value of the token for a target element.
     */
    deleteValueFor(target) {
        FASTDesignTokenNode.getOrCreate(target).deleteTokenValue(this);
        return this;
    }
    /**
     * Sets the default value of the token.
     */
    withDefault(value) {
        FASTDesignTokenNode.defaultNode.setTokenValue(this, this.normalizeValue(value));
        return this;
    }
    /**
     * Subscribes a subscriber to notifications for the token.
     */
    subscribe(subscriber) {
        this.subscribers.subscribe(subscriber);
    }
    /**
     * Unsubscribes a subscriber to notifications for the token.
     */
    unsubscribe(subscriber) {
        this.subscribers.unsubscribe(subscriber);
    }
    /**
     * Alias the token to the provided token.
     * @param token - the token to alias to
     */
    alias(token) {
        return ((resolve) => resolve(token));
    }
    normalizeValue(value) {
        if (value instanceof DesignToken) {
            value = this.alias(value);
        }
        return value;
    }
}
/**
 * @public
 */
let CSSDesignToken = class CSSDesignToken extends DesignToken {
    constructor(configuration) {
        super(configuration);
        this.cssReflector = {
            handleChange: (source, record) => {
                const target = record.target === FASTDesignTokenNode.defaultNode
                    ? FASTDesignTokenNode.rootStyleSheetTarget
                    : record.target instanceof FASTDesignTokenNode
                        ? PropertyTargetManager.getOrCreate(record.target.target)
                        : null;
                if (target) {
                    if (record.type === 2 /* DesignTokenMutationType.delete */) {
                        target.removeProperty(this.cssCustomProperty);
                    }
                    else {
                        target.setProperty(this.cssCustomProperty, this.resolveCSSValue(record.target.getTokenValue(this)));
                    }
                }
            },
        };
        this.cssCustomProperty = `--${configuration.cssCustomPropertyName}`;
        this.cssVar = `var(${this.cssCustomProperty})`;
        Observable.getNotifier(this).subscribe(this.cssReflector);
    }
    /**
     * The DesignToken represented as a string that can be used in CSS.
     */
    createCSS() {
        return this.cssVar;
    }
    resolveCSSValue(value) {
        return value && typeof value.createCSS === "function" ? value.createCSS() : value;
    }
};
CSSDesignToken = __decorate([
    cssDirective()
], CSSDesignToken);
export { CSSDesignToken };
const defaultDesignTokenResolutionStrategy = {
    contains: composedContains,
    parent(element) {
        let parent = composedParent(element);
        while (parent !== null) {
            if (parent instanceof FASTElement) {
                return parent;
            }
            parent = composedParent(parent);
        }
        return null;
    },
    bind() { },
    unbind() { },
};
class FASTDesignTokenNode extends DesignTokenNode {
    constructor(target) {
        super();
        this.target = target;
    }
    static get strategy() {
        if (this._strategy === undefined) {
            FASTDesignTokenNode.withStrategy(defaultDesignTokenResolutionStrategy);
        }
        return this._strategy;
    }
    bind(target) {
        let parent = FASTDesignTokenNode.findParent(target);
        if (parent === null) {
            parent = FASTDesignTokenNode.defaultNode;
        }
        if (parent !== this.parent) {
            const reparent = [];
            for (const child of parent.children) {
                if (child instanceof FASTDesignTokenNode &&
                    FASTDesignTokenNode.strategy.contains(target, child.target)) {
                    reparent.push(child);
                }
            }
            parent.appendChild(this);
            for (const child of reparent) {
                this.appendChild(child);
            }
        }
    }
    unbind() {
        FASTDesignTokenNode.cache.delete(this.target);
        this.dispose();
    }
    static getOrCreate(target) {
        let found = FASTDesignTokenNode.cache.get(target);
        if (found) {
            return found;
        }
        found = new FASTDesignTokenNode(target);
        FASTDesignTokenNode.cache.set(target, found);
        target.$fastController.addBehaviors([FASTDesignTokenNode.strategy, found]);
        return found;
    }
    static withStrategy(strategy) {
        this._strategy = strategy;
    }
    static findParent(target) {
        let current = FASTDesignTokenNode.strategy.parent(target);
        while (current !== null) {
            const node = FASTDesignTokenNode.cache.get(current);
            if (node) {
                return node;
            }
            current = FASTDesignTokenNode.strategy.parent(current);
        }
        return null;
    }
}
FASTDesignTokenNode.defaultNode = new DesignTokenNode();
FASTDesignTokenNode.rootStyleSheetTarget = new RootStyleSheetTarget();
FASTDesignTokenNode.cache = new WeakMap();
